from typed_ast.ast3 import AnnAssign, Assign, Index, Load, Module, Name, Store, Str, Subscript

from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import is_valid_identifier, is_valid_type
from arcor2.source.utils import find_function, get_name_attr


def object_instance_from_res(tree: Module, object_name: str, object_id: str, cls_name: str) -> None:

    if not is_valid_identifier(object_name):
        raise Arcor2Exception(f"Object name {object_name} is not a valid identifier.")

    if not is_valid_type(cls_name):
        raise Arcor2Exception(f"Class name {cls_name} is not valid.")

    main_body = find_function("main", tree).body
    last_assign_idx = -1

    for body_idx, body_item in enumerate(main_body):

        if isinstance(body_item, (Assign, AnnAssign)):
            last_assign_idx = body_idx

    assign = AnnAssign(
        target=Name(id=object_name, ctx=Store()),
        annotation=Name(id=cls_name, ctx=Load()),
        value=Subscript(
            value=get_name_attr("res", "objects"), slice=Index(value=Str(s=object_id, kind="")), ctx=Load()
        ),
        simple=1,
    )

    main_body.insert(last_assign_idx + 1, assign)



#!/usr/bin/env python3

import argparse
import json
import logging
import os
import zipfile
from datetime import datetime, timezone
from io import BytesIO
from typing import Set, Tuple, Union

import humps  # type: ignore
from apispec import APISpec  # type: ignore
from apispec_webframeworks.flask import FlaskPlugin  # type: ignore
from flask import Flask, Response, request, send_file
from flask_cors import CORS  # type: ignore
from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore

import arcor2_build
from arcor2.cached import CachedProject, CachedScene
from arcor2.clients import persistent_storage as ps
from arcor2.data.execution import PackageMeta
from arcor2.data.object_type import ObjectModel, ObjectType
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import port_from_url
from arcor2.logging import get_logger
from arcor2.object_types.utils import base_from_source, built_in_types_names
from arcor2.source import SourceException
from arcor2.source.utils import parse
from arcor2_build.source.logic import program_src
from arcor2_build.source.utils import derived_resources_class, global_action_points_class, global_actions_class
from arcor2_build_data import SERVICE_NAME, URL

logger = get_logger("Build")

# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2_build.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin()],
)

app = Flask(__name__)
CORS(app)

RETURN_TYPE = Union[Tuple[str, int], Response]


def get_base(
    object_types: Set[str], written_types: Set[str], obj_type: ObjectType, zf: zipfile.ZipFile, ot_path: str
) -> None:

    base = base_from_source(obj_type.source, obj_type.id)

    if not base:
        return

    if base in written_types:
        return

    if base not in object_types and base not in built_in_types_names():

        logger.debug(f"Getting {base} as base of {obj_type.id}.")
        base_obj_type = ps.get_object_type(base)

        zf.writestr(os.path.join(ot_path, humps.depascalize(base_obj_type.id)) + ".py", base_obj_type.source)
        object_types.add(base)
        written_types.add(base)

        # try to get base of the base
        get_base(object_types, written_types, base_obj_type, zf, ot_path)


def _publish(project_id: str, package_name: str) -> RETURN_TYPE:

    mem_zip = BytesIO()

    logger.debug(f"Generating package {package_name} for project_id: {project_id}.")

    with zipfile.ZipFile(mem_zip, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:

        try:
            logger.debug("Getting scene and project.")
            project = ps.get_project(project_id)
            cached_project = CachedProject(project)
            scene = ps.get_scene(project.scene_id)
            cached_scene = CachedScene(scene)

            data_path = "data"
            ot_path = "object_types"

            zf.writestr(os.path.join(ot_path, "__init__.py"), "")
            zf.writestr(os.path.join(data_path, "project.json"), project.to_json())
            zf.writestr(os.path.join(data_path, "scene.json"), scene.to_json())

            obj_types = set(cached_scene.object_types)
            obj_types_with_models: Set[str] = set()
            written_types: Set[str] = set()

            for scene_obj in scene.objects:

                if scene_obj.type in written_types:
                    continue

                logger.debug(f"Getting scene object type {scene_obj.type}.")
                obj_type = ps.get_object_type(scene_obj.type)

                if obj_type.model and obj_type.id not in obj_types_with_models:
                    obj_types_with_models.add(obj_type.id)

                    model = ps.get_model(obj_type.model.id, obj_type.model.type)
                    obj_model = ObjectModel(obj_type.model.type, **{model.type().value.lower(): model})  # type: ignore

                    zf.writestr(
                        os.path.join(data_path, "models", humps.depascalize(obj_type.id) + ".json"), obj_model.to_json()
                    )

                zf.writestr(os.path.join(ot_path, humps.depascalize(obj_type.id)) + ".py", obj_type.source)
                written_types.add(scene_obj.type)

                # handle inheritance
                get_base(obj_types, written_types, obj_type, zf, ot_path)

        except Arcor2Exception as e:
            logger.exception("Failed to get something from the project service.")
            return str(e), 404

        script_path = "script.py"

        try:

            if project.has_logic:
                logger.debug("Generating script from project logic.")
                zf.writestr(script_path, program_src(cached_project, cached_scene, built_in_types_names(), True))
            else:
                try:
                    logger.debug("Getting project sources.")
                    script = ps.get_project_sources(project.id).script

                    # check if it is a valid Python code
                    try:
                        parse(script)
                    except SourceException:
                        logger.exception("Failed to parse code of the uploaded script.")
                        return "Invalid code.", 501

                    zf.writestr(script_path, script)

                except ps.ProjectServiceException:

                    logger.info("Script not found on project service, creating one from scratch.")

                    # write script without the main loop
                    zf.writestr(script_path, program_src(cached_project, cached_scene, built_in_types_names(), False))

            logger.debug("Generating supplementary files.")

            logger.debug("resources.py")
            zf.writestr("resources.py", derived_resources_class(cached_project))

            logger.debug("actions.py")
            zf.writestr("actions.py", global_actions_class(cached_project))

            logger.debug("action_points.py")
            zf.writestr("action_points.py", global_action_points_class(cached_project))

            logger.debug("package.json")
            zf.writestr("package.json", PackageMeta(package_name, datetime.now(tz=timezone.utc)).to_json())

        except Arcor2Exception as e:
            logger.exception("Failed to generate script.")
            return str(e), 501

    logger.debug("Execution package finished.")
    mem_zip.seek(0)
    return send_file(mem_zip, as_attachment=True, cache_timeout=0, attachment_filename="arcor2_project.zip")


@app.route("/project/<string:project_id>/publish", methods=["GET"])
def project_publish(project_id: str) -> RETURN_TYPE:
    """Publish project
    ---
    get:
      description: Get zip file with execution package. To be used by the Execution service.
      parameters:
        - in: path
          name: project_id
          schema:
            type: string
          required: true
          description: unique ID
        - in: query
          name: packageName
          schema:
            type: string
            default: N/A
          required: false
          description: Package name
      responses:
        200:
          description: Ok
          content:
            application/zip:
                schema:
                  type: string
                  format: binary
                  example: The archive of execution package (.zip)
        404:
            description: Project ID or some of the required items was not found.
        501:
            description: Project invalid.
    """

    return _publish(project_id, request.args.get("packageName", default="N/A"))


@app.route("/project/<string:project_id>/script", methods=["PUT"])
def project_script(project_id: str):
    """Project script
    ---
    put:
      description: Add or update project main script (DOES NOT WORK YET).
      parameters:
        - in: path
          name: project_id
          schema:
            type: string
          required: true
          description: unique ID
      requestBody:
          content:
            text/x-python:
              schema:
                type: string
                format: binary
      responses:
        200:
          description: Ok
    """
    # TODO use get_logic_from_source
    pass


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger() -> str:
    return json.dumps(spec.to_dict())


with app.test_request_context():
    spec.path(view=project_publish)
    spec.path(view=project_script)


def main() -> None:

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument("-s", "--swagger", action="store_true", default=False)
    parser.add_argument(
        "-d",
        "--debug",
        help="Set logging level to debug.",
        action="store_const",
        const=logging.DEBUG,
        default=logging.INFO,
    )

    args = parser.parse_args()
    logger.setLevel(args.debug)

    if args.swagger:
        print(spec.to_yaml())
        return

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL, "./api/swagger.json"  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host="0.0.0.0", port=port_from_url(URL))


if __name__ == "__main__":
    main()



import re
from typing import Any, Dict, List, Optional

from typed_ast.ast3 import (
    Assign,
    Attribute,
    Call,
    ClassDef,
    Compare,
    Eq,
    ExceptHandler,
    Expr,
    FunctionDef,
    If,
    ImportFrom,
    Index,
    Load,
    Module,
    Name,
    NameConstant,
    NodeTransformer,
    NodeVisitor,
    Pass,
    Return,
    Store,
    Str,
    Subscript,
    Try,
    While,
    With,
    alias,
    arg,
    arguments,
    keyword,
    stmt,
    withitem,
)

import arcor2.data.common
import arcor2.exceptions.runtime
from arcor2.cached import CachedProject
from arcor2.data.common import ActionPoint
from arcor2.source import SourceException
from arcor2.source.utils import add_import, find_function, get_name_attr, parse, tree_to_str


def main_loop_body(tree: Module) -> List[Any]:
    main = find_function("main", tree)

    for node in main.body:
        if isinstance(node, While):  # TODO more specific condition (test for True argument)
            return node.body

    raise SourceException("Main loop not found.")


def empty_script_tree(add_main_loop: bool = True) -> Module:
    """Creates barebones of the script (empty 'main' function).

    Returns
    -------
    """

    main_body: List[stmt] = []

    if add_main_loop:
        main_body.append(While(test=NameConstant(value=True), body=[Pass()], orelse=[]))
    else:
        """put there "pass" in order to make code valid even if there is no
        other statement (e.g. no object from resources)"""
        main_body.append(Pass())

    # TODO helper function for try ... except

    tree = Module(
        body=[
            FunctionDef(
                name="main",
                args=arguments(
                    args=[arg(arg="res", annotation=Name(id="Resources", ctx=Load()), type_comment=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[],
                ),
                body=main_body,
                decorator_list=[],
                returns=NameConstant(value=None),
                type_comment=None,
            ),
            If(
                test=Compare(
                    left=Name(id="__name__", ctx=Load()), ops=[Eq()], comparators=[Str(s="__main__", kind="")]
                ),
                body=[
                    Try(
                        body=[
                            With(
                                items=[
                                    withitem(
                                        context_expr=Call(func=Name(id="Resources", ctx=Load()), args=[], keywords=[]),
                                        optional_vars=Name(id="res", ctx=Store()),
                                    )
                                ],
                                body=[
                                    Expr(
                                        value=Call(
                                            func=Name(id="main", ctx=Load()),
                                            args=[Name(id="res", ctx=Load())],
                                            keywords=[],
                                        )
                                    )
                                ],
                                type_comment=None,
                            )
                        ],
                        handlers=[
                            ExceptHandler(
                                type=Name(id="Exception", ctx=Load()),
                                name="e",
                                body=[
                                    Expr(
                                        value=Call(
                                            func=Name(id="print_exception", ctx=Load()),
                                            args=[Name(id="e", ctx=Load())],
                                            keywords=[],
                                        )
                                    )
                                ],
                            )
                        ],
                        orelse=[],
                        finalbody=[],
                    )
                ],
                orelse=[],
            ),
        ],
        type_ignores=[],
    )

    add_import(tree, arcor2.exceptions.runtime.__name__, arcor2.exceptions.runtime.print_exception.__name__)
    add_import(tree, "resources", "Resources", try_to_import=False)

    return tree


def add_cls_inst(
    node: Module, cls: str, name: str, kwargs: Optional[Dict] = None, kwargs2parse: Optional[Dict] = None
) -> None:
    class FindImport(NodeVisitor):
        def __init__(self, cls: str) -> None:

            self.found = False
            self.cls = cls

        def visit_ImportFrom(self, node: ImportFrom) -> None:

            for node_alias in node.names:
                if node_alias.name == self.cls:
                    self.found = True

            if not self.found:
                self.generic_visit(node)

    class FindClsInst(NodeVisitor):
        def __init__(self) -> None:

            self.found = False

        def visit_FunctionDef(self, node: FunctionDef) -> None:

            if node.name == "main":

                for item in node.body:

                    if isinstance(item, Assign):

                        assert isinstance(item.targets[0], Name)

                        if item.targets[0].id == name:

                            # TODO assert for item.value
                            if item.value.func.id != cls:  # type: ignore
                                raise SourceException(
                                    f"Name {name} already used for instance of {item.value.func.id}!"  # type: ignore
                                )

                            self.found = True
                            # TODO update arguments?

            if not self.found:
                self.generic_visit(node)

    class AddClsInst(NodeTransformer):
        def visit_FunctionDef(self, node: FunctionDef) -> FunctionDef:

            if node.name == "main":

                kw = []

                if kwargs:
                    for k, v in kwargs.items():
                        kw.append(keyword(arg=k, value=v))

                if kwargs2parse:
                    for k, v in kwargs2parse.items():
                        kw.append(keyword(arg=k, value=parse(v)))

                node.body.insert(
                    0,
                    Assign(
                        targets=[Name(id=name, ctx=Store())],
                        value=Call(func=Name(id=cls, ctx=Load()), args=[], keywords=kw),
                    ),
                )

            return node

    find_import = FindImport(cls)
    find_import.visit(node)

    if not find_import.found:
        raise SourceException("Class '{}' not imported!".format(cls))

    vis = FindClsInst()
    vis.visit(node)

    if not vis.found:
        tr = AddClsInst()
        node = tr.visit(node)


def add_method_call_in_main(tree: Module, instance: str, method: str, args: List, kwargs: List) -> None:
    """Places method call after block where instances are created.

    Parameters
    ----------
    tree
    instance
    init
    args
    kwargs

    Returns
    -------
    """

    main_body = find_function("main", tree).body
    last_assign_idx = None

    for body_idx, body_item in enumerate(main_body):

        # TODO check if instance exists!

        if isinstance(body_item, Assign):
            last_assign_idx = body_idx

    if not last_assign_idx:
        raise SourceException()

    # TODO iterate over args/kwargs
    # TODO check actual number of method's arguments (and types?)
    main_body.insert(
        last_assign_idx + 1, Expr(value=Call(func=get_name_attr(instance, method), args=args, keywords=[]))
    )


def clean(x: str) -> str:
    return re.sub("\W|^(?=\d)", "_", x)  # noqa


def global_action_points_class(project: CachedProject) -> str:
    tree = Module(body=[])
    tree.body.append(
        ImportFrom(module=arcor2.data.common.__name__, names=[alias(name=ActionPoint.__name__, asname=None)], level=0)
    )
    tree.body.append(ImportFrom(module="resources", names=[alias(name="Resources", asname=None)], level=0))

    cls_def = ClassDef(
        name="ActionPoints",
        bases=[],
        keywords=[],
        body=[
            FunctionDef(
                name="__init__",
                args=arguments(
                    args=[
                        arg(arg="self", annotation=None, type_comment=None),
                        arg(arg="res", annotation=Name(id="Resources", ctx=Load()), type_comment=None),
                    ],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[],
                ),
                body=[
                    Assign(
                        targets=[Attribute(value=Name(id="self", ctx=Load()), attr="_res", ctx=Store())],
                        value=Name(id="res", ctx=Load()),
                        type_comment=None,
                    )
                ],
                decorator_list=[],
                returns=None,
                type_comment=None,
            )
        ],
        decorator_list=[],
    )

    for ap in project.action_points:
        fd = FunctionDef(
            name=clean(ap.name),  # TODO avoid possible collisions
            args=arguments(
                args=[arg(arg="self", annotation=None, type_comment=None)],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[],
            ),
            body=[
                Return(
                    value=Call(
                        func=Attribute(
                            value=Attribute(
                                value=Attribute(value=Name(id="self", ctx=Load()), attr="_res", ctx=Load()),
                                attr="project",
                                ctx=Load(),
                            ),
                            attr="action_point",
                            ctx=Load(),
                        ),
                        args=[Str(s=ap.id, kind="")],
                        keywords=[],
                    )
                )
            ],
            decorator_list=[Name(id="property", ctx=Load())],
            returns=Name(id="ActionPoint", ctx=Load()),
            type_comment=None,
        )

        cls_def.body.append(fd)

    tree.body.append(cls_def)
    return tree_to_str(tree)


def global_actions_class(project: CachedProject) -> str:
    tree = Module(body=[])
    tree.body.append(ImportFrom(module="resources", names=[alias(name="Resources", asname=None)], level=0))

    cls_def = ClassDef(
        name="Actions",
        bases=[],
        keywords=[],
        body=[
            FunctionDef(
                name="__init__",
                args=arguments(
                    args=[
                        arg(arg="self", annotation=None, type_comment=None),
                        arg(arg="res", annotation=Name(id="Resources", ctx=Load()), type_comment=None),
                    ],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[],
                ),
                body=[
                    Assign(
                        targets=[Attribute(value=Name(id="self", ctx=Load()), attr="_res", ctx=Store())],
                        value=Name(id="res", ctx=Load()),
                        type_comment=None,
                    )
                ],
                decorator_list=[],
                returns=None,
                type_comment=None,
            )
        ],
        decorator_list=[],
    )

    for action in project.actions:

        ac_obj, ac_type = action.parse_type()

        m = FunctionDef(
            name=clean(action.name),
            args=arguments(
                args=[arg(arg="self", annotation=None, type_comment=None)],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[],
            ),
            body=[
                Expr(
                    value=Call(
                        func=Attribute(
                            value=Subscript(
                                value=Attribute(
                                    value=Attribute(value=Name(id="self", ctx=Load()), attr="_res", ctx=Load()),
                                    attr="all_instances",
                                    ctx=Load(),
                                ),
                                slice=Index(value=Str(s=ac_obj, kind="")),
                                ctx=Load(),
                            ),
                            attr=ac_type,
                            ctx=Load(),
                        ),
                        args=[
                            Attribute(
                                value=Attribute(value=Name(id="self", ctx=Load()), attr="_res", ctx=Load()),
                                attr=clean(action.name),
                                ctx=Load(),
                            )
                        ],
                        keywords=[],
                    )
                )
            ],
            decorator_list=[],
            returns=None,
            type_comment=None,
        )

        cls_def.body.append(m)

    tree.body.append(cls_def)
    return tree_to_str(tree)


def derived_resources_class(project: CachedProject) -> str:

    tree = Module(body=[])

    parameters = [(act.id, clean(act.name)) for act in project.actions]

    # TODO rather import resources/ResourcesBase and get name from __name__ (but it requires ARCOR2_PROJECT_PATH...)
    add_import(tree, "arcor2.resources", "ResourcesBase", try_to_import=False)

    derived_cls_name = "Resources"

    init_body: List = [
        Expr(
            value=Call(
                func=Attribute(
                    value=Call(
                        func=Name(id="super", ctx=Load()),
                        args=[Name(id=derived_cls_name, ctx=Load()), Name(id="self", ctx=Load())],
                        keywords=[],
                    ),
                    attr="__init__",
                    ctx=Load(),
                ),
                args=[Str(s=project.id)],
                keywords=[],
            )
        )
    ]

    for a_id, a_name in parameters:
        init_body.append(
            Assign(
                targets=[get_name_attr("self", "_" + a_name, Store)],
                value=Call(func=get_name_attr("self", "parameters"), args=[Str(s=a_id)], keywords=[]),
            )
        )

    cls_def = ClassDef(
        name=derived_cls_name,
        bases=[Name(id="ResourcesBase", ctx=Load())],
        keywords=[],
        body=[
            FunctionDef(
                name="__init__",
                args=arguments(
                    args=[arg(arg="self", annotation=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[],
                ),
                body=init_body,
                decorator_list=[],
                returns=None,
            )
        ],
        decorator_list=[],
    )

    tree.body.append(cls_def)

    for a_id, a_name in parameters:
        cls_def.body.append(
            FunctionDef(
                name=a_name,
                args=arguments(
                    args=[arg(arg="self", annotation=None, type_comment=None)],
                    vararg=None,
                    kwonlyargs=[],
                    kw_defaults=[],
                    kwarg=None,
                    defaults=[],
                ),
                body=[
                    Expr(
                        value=Call(
                            func=Attribute(value=Name(id="self", ctx=Load()), attr="print_info", ctx=Load()),
                            args=[Str(s=a_id, kind=""), get_name_attr("self", "_" + a_name)],
                            keywords=[],
                        )
                    ),
                    Return(value=get_name_attr("self", "_" + a_name)),
                ],
                decorator_list=[Name(id="property", ctx=Load())],
                returns=None,
                type_comment=None,
            )
        )

    return tree_to_str(tree)
